/* Original class by David Lovell drl@eng.cam.ac.uk.
   Subsequent changes by Steve Waterhouse srw1001@eng.cam.ac.uk
*/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "incs.h"
#include "conj.h"

static const int ErrBufSize = 256;
static char ErrBuf[ErrBufSize];

// Flag to indicate whether control is currently within the linesearch
// of the conjugate gradient optimization routine. Since it is
// expected that SQRT: domain errors will arise in this routine as
// part of the normal running of the optimizer, this flag allows
// matherr to process the exception without writing to stderr and
// clogging up the screen.
static int InLNS = 0;

// The linesearch of the conjugate gradient optimizer makes lots of
// complaints about sqrt() domain errors. This exception handler stops
// them from clogging up the display.

/* --------------------------------------------------------------------------
   Adaptation of conjugate gradient minimiser by Carl Edward
   Rasmussen, available at http://www.cs.utoronto.ca/~delve/

   The original source was written in C. I have tried to port it to
   C++ with minimal changes. Here are a list of the alterations:
   
   extern int  no_wts;            -> Conj::NumWts
   extern real *w;                -> member variable Conj::W
   extern real fgeval(real *dw)   -> virtual double evaluate_gradient(double *dw) = 0;
   real *dw1, *dw2, *s;           -> double *dw1, *dw2, *s;
   int nfgeval;                   -> local to conj, passed to lns as *nfgeval
   real                           -> double
   extern int  TIMEOUT;           -> not used

   Also the variable InLNS is set to 1 when control is inside the
   Conj::lns() routine. This allows the matherr exception handler to
   absorb the SQRT: domain errors which arise naturally.
   ------------------------------------------------------------------------- */

// *f1                current function value
// *z                 guess for initial step
// d1                 slope
// *nfgeval           number of func/grad evals so far; instead of this
//                    being incremented by fgeval(), it is incremented
//                    in lns() each time evaluate_gradient() is called.

int Conj::lns(double *f1, double *z, double d1, int *nfgeval){
  InLNS = 1;
  double RHO = 0.25, SIG = 0.5, INT = 0.1, EXT = 3.0;
  int  MAX = 20;
  double d2, d3, f2, f3, z2, z3, A, B, max = -1.0;
  int  i, k;

  for (i=0; i<NumWts; i++) 
    W[i] += *z*s[i];                // update weights 
  f2 = evaluate_gradient(dw2); 
  (*nfgeval)++;

  d2 = 0.0; 
  for (i=0; i<NumWts; i++) 
    d2 += dw2[i]*s[i];

  f3=*f1; 
  d3=d1; 
  z3=-*z;              // initialize point 3 equal to point 1 

  k = *nfgeval + MAX;
  while (*nfgeval < k) {     
    // allow limited amount of search 
    while (((f2 > *f1+*z*RHO*d1) || (d2 > -SIG*d1)) && (*nfgeval < k)) {
      max=*z;                                  
      // tighten the bracket 
      if (f2 > *f1) z2=z3-(0.5*d3*z3*z3)/(d3*z3+f2-f3); 
      // quadratic fit 
      else {                                                 
	// cubic fit 
        A = 6.0*(f2-f3)/z3+3.0*(d2+d3);
        B = 3.0*(f3-f2)-z3*(d3+2.0*d2);
        z2 = (sqrt(B*B-A*d2*z3*z3)-B)/A;  
	// numerical error possible - ok! 
      }
      if (z2 != z2) 
	z2 = z3/2;                
      // if z2 is NaN then bisection 
      if (z2 > INT*z3) 
	z2 = INT*z3;      
      // bound solution away from current 
      if (z2 < (1.0-INT)*z3) 
	z2 = (1.0-INT)*z3;        // bound away from z3 
      *z += z2;                   // update absolute stepsize 
      for (i=0; i<NumWts; i++) 
	W[i] += z2*s[i];            
      // update weights 
      f2 = evaluate_gradient(dw2); 
      (*nfgeval)++;
      d2 = 0.0; 
      for (i=0; i<NumWts; i++) 
	d2 += dw2[i]*s[i];
      z3 -= z2;
      // z3 is now relative to the location of z2 
    }

    if (d2 > SIG*d1) { 
      *f1 = f2; 
      InLNS = 0; 
      return 1;       // SUCCESS 
    }     
      
    else {
      // DO SOMETHING HERE !!!
    }

    A = 6.0*(f2-f3)/z3+3.0*(d2+d3); 
    // make cubic extrapolation 
    B = 3.0*(f3-f2)-z3*(d3+2.0*d2);
    z2 = -d2*z3*z3/(B+sqrt(B*B-A*d2*z3*z3));
    // num. error possible - ok! 
    if (z2 != z2)                                            
      // z2 is NaN? 
      z2 = (max < -0.5) ? *z*(EXT-1.0) : 0.5*(max-*z);      
    // bisection 
    else if (z2 < 0.0)              
      // minimum is to the left of current? 
      z2 = (max < -0.5) ? *z*(EXT-1.0) : 0.5*(max-*z);         
    // bisection 
    else if ((max > -0.5) && (z2+*z > max)) 
      // extrap. beyond max? 
      z2 = 0.5*(max-*z);                                      
    // bisection 
    else if ((max < -0.5) && (z2+*z > *z*EXT))     
      // extrap. beyond EXT? 
      z2 = *z*(EXT-1.0);                          
    // set to extrap. limit 
    else if (z2<-z3*INT)    
      // too close to current point? 
      z2 = -z3*INT;
    else if ((max > -0.5) && (z2 < (max-*z)*(1.0-INT))) 
      // too close to max? 
      z2 = (max-*z)*(1.0-INT);

    // swap point 2 and 3 
    f3=f2; 
    d3=d2; 
    z3=-z2;       

    *z += z2;

    // update weights 
    for (i=0; i<NumWts; i++) 
      W[i] += z2*s[i];           

    f2 = evaluate_gradient(dw2);
    (*nfgeval)++;
    d2 = 0.0; 
    for (i=0; i<NumWts; i++) 
      d2 += dw2[i]*s[i];
  }
  *f1 = f2;
  InLNS = 0;
  return 0;                                             // linesearch failed 
}

// *iter              max iterations allowed. Actual number is returned.
// *epoch             max epochs allowed. Actual number is returned.
// *restart           if (*restart) then restart CG with steepest descent
// *costvalue         returns the value of the costfunction

int Conj::optimise(int *iter, int *epoch, int *restart, double *costvalue){

  if(!dw1) CreateConj();      // allocate working memory if not already done 
  
  static int    ls_failed = 0;             // set to 1 if line search failed 
  int           j,                         // miscellaneous counter 
                cur_iter = 0;              // counts current iteration 
  static double fun, slope1, step;
  double        *tmp, y, z, q, slope2;

  int nfgeval = 0; // number of function and gradient evaluations 
  if (*restart) {  
    // start by using direction of steepest descent 
    //    printf("using steepest descent...\n");
    fun = evaluate_gradient(dw1);  nfgeval++;
    slope1 = 0.0;
    for (j=0; j<NumWts; j++) {
      s[j] = -dw1[j]; slope1 -= s[j]*s[j];
    }
    // set initial step-size to 1/(|s|+1) 
    step = -1.0/(slope1-1.0);    

    // probably we won't want to restart on next call 
    *restart = 0;       
  }
  while (((*epoch == 0) || (nfgeval < *epoch)) &&
         ((*iter == 0) || (cur_iter < *iter))) {  
    cur_iter++;
    if (lns(&fun, &step, slope1, &nfgeval)) {
      // if line search succeeded 
      //      printf("line search succeeded\n");
      y = z = q = 0.0; 
      for (j=0; j<NumWts; j++){ 
	y += dw2[j] * dw2[j]; 
	z += dw1[j] * dw2[j]; 
	q += dw1[j] * dw1[j]; 
      }
      y = (y-z)/q; 
      for (j=0; j<NumWts; j++) 
	s[j] = y*s[j]-dw2[j];        // new direction 

      // swap derivatives 
      tmp = dw2; 
      dw2 = dw1; 
      dw1 = tmp;    

      slope2 = 0.0; 
      for (j=0; j<NumWts; j++) 
	slope2 += dw1[j]*s[j];
      if (slope2 > 0.0) {  
	// must be negative, else use steepest descent 
	slope2 = 0.0;
	for (j=0; j<NumWts; j++) { 
	  s[j] = -dw1[j]; 
	  slope2 -= s[j]*s[j]; 
	}
      } 
      step *= (slope1/slope2 > 100.0) ? 100.0 : slope1/slope2; 
      slope1 = slope2;
      ls_failed = 0; 
    } 
    else {                             // line search failed 
      //      printf("line search failed\n");
      if (ls_failed) {
	// break if previous failed, else try steepest 
	*epoch = nfgeval; 
	*iter = cur_iter; 
	*costvalue = fun; 
	return 0; 
      }
      // swap derivatives 
      dw1 = dw2;   
      slope1 = 0.0;
      for (j=0; j<NumWts; j++) { 
	s[j] = -dw1[j]; 
	slope1 -= s[j]*s[j]; 
      }
      // set new step-size guess to 1/(|s|+1) 
      step = -1.0/(slope1-1.0);   
      ls_failed = 1; 
    }
  }
  *epoch = nfgeval; 
  *iter = cur_iter; 
  *costvalue = fun; 
  return 1; 
}

void Conj::CreateConj(){
  dw1 = new double[NumWts];
  dw2 = new double[NumWts];
  s   = new double[NumWts];
}

// --------------------------------------------------------------------------

Conj::Conj(){
  NumWts = 0;
  W      = NULL;
  dw1    = NULL;
  dw2    = NULL;
  s      = NULL;
}

void Conj::init(int NumWts){
  if(NumWts <= 0){
    printf("Attempt to initialize a Conj object with NumWts <= 0. Exiting.\n");
    exit(0);	
  }
  Conj::NumWts = NumWts;
  W = new double[Conj::NumWts];
}
  
Conj::Conj(int NumWts){
  init(NumWts);
  dw1    = NULL;
  dw2    = NULL;
  s      = NULL;
}

Conj::~Conj(){
  if(W)   delete [] W;
  if(dw1) delete [] dw1;
  if(dw2) delete [] dw2;
  if(s)   delete [] s;
}

// Save the weights to specified file. Return 1 if all weights were
// sucessfully saved. Return 0 if not all weights could be
// stored.
int Conj::SaveWeights(const char *outfile) const{
  FILE *out;

  if(!(out = fopen(outfile, "w"))){
	sprintf(ErrBuf, "Could not open \"%s\" in SaveWeights()\n", outfile);
	printf("%s", ErrBuf);
	return 0;
  }
  
  int NumSaved = fwrite((const void *) W, sizeof(double), NumWts, out);
  fclose(out);
  return (NumSaved == NumWts ? 1 : 0);
}

// Load the weights from specified file. Return 1 if all weights were
// sucessfully loaded. Return 0 if not all weights could be
// loaded. Return -1 if there were additional weights still to be
// loaded from the given file.
int Conj::LoadWeights(const char *infile) const{
  FILE *in;
 
  if(!(in = fopen(infile, "r"))){
	sprintf(ErrBuf, "Could not open \"%s\" in LoadWeights()\n", infile);
	printf("%s", ErrBuf);
	return 0;
  }
  
  int NumLoaded = fread((void *) W,	sizeof(double), NumWts, in);

  int state;
  double test;
  if(fread((void *) &test, sizeof(double), 1, in) == 1)
	state = -1;
  else
	state = (NumLoaded == NumWts ? 1 : 0);
  
  fclose(in);
  return state;
}
