#ifndef CONJ_H
#define CONJ_H
/* --------------------------------------------------------------------------
  This is a base class from which specific functions can be
  derived. The only information that we need to know about the
  function at this level is how many parameters it has. Since this
  class was originally developed with neural networks in mind, the
  parameters of the function are called "weights". The idea behind
  this class is to provide a framework for function minimization. In
  this header are declarations for a number of function minimization
  techniques which require the evaluation of the function and/or its
  gradient with respect to its parameters. The implementation of this
  evaluation is left up to classes derived from this base.
  ------------------------------------------------------------------------- */

/* Original class by David Lovell drl@eng.cam.ac.uk.
   Subsequent changes by Steve Waterhouse srw1001@eng.cam.ac.uk
*/


class Conj{

public:
  Conj(int NumWts);
  virtual ~Conj() = 0;

  int optimise(int *iter, int *epoch, int *restart, double *costvalue);

  // "optimise" is an adaptation of the conjugate gradient minimiser written
  // by Carl Edward Rasmussen, available at http://www.cs.utoronto.ca/~delve/.
  //
  // iter      is the maximum number of epochs allowed. The actual number
  //           is returned.
  // epoch     is the maximum number of epochs allowed. The actual number
  //           is returned.
  //
  // if (*restart) then restart CG with steepest descent
  //
  // costvalue returns the value of the costfunction 
  // (ie. evaluate_gradient() above)


 
protected:

  // ok, this is the set of members and functions that must
  // be defined in the derived class.

  double *W;			
  // The weights - set this to point
  // to your weights in the derived class

  int NumWts;
  // the number of weights - this is set in the constructor

  // Evaluate the function and its gradient with respect to the
  // weights. "evaluate_gradient" returns the value of the function and sets
  // dw, an array of NumWts doubles, to the gradient of the function
  // with respect to the weights. "evaluate_gradient" is used by the conjugate
  // gradient minimiser and *MUST* be defined in the class derived from
  // this base class.

  virtual double evaluate_gradient(double *dw) = 0;
  
  // This constructor and the init() function are to be used by
  // derived classes that read their initialization from an init file.
  // They dont need to be considered if you dont bother using an init
  // file. 

  Conj();
  void init(int NumWts);

public: 
  
  // Save/load the weights to/from specified file. Return 1 if all weights
  // were sucessfully saved/loaded. Return 0 if not all weights could be
  // stored/loaded. Return -1 if there were additional weights still to 
  // be loaded from the given file.

  int SaveWeights(const char *outfile) const;
  int LoadWeights(const char *infile) const;

  // Save/load the weights to/from specified array of doubles

  void SaveWeights(double *wptr) const{
	for(int i = 0; i < NumWts; i++) wptr[i] = W[i];
  }

  void LoadWeights(double *wptr) const{
	for(int i = 0; i < NumWts; i++) W[i] = wptr[i];
  }

  // Return  the number of weights.
  int NumWeights() const{ return NumWts; }

  // (fn) is true if it has been initialized correctly
  operator void* () const{return (void *) (NumWts != 0);}

  // !fn is false if it has been initialized correctly
  int operator ! () const{return (NumWts == 0);}

private:
  // --------------------------------------------------------------------------
  // These functions and variables are used in the conjugate gradient
  // minimization routines
  double *dw1, *dw2;	// Arrays of derivatives for all weights
  double *s;			// search direction used for linesearches
 
  // Line search used by the public member conj()
  int lns(double *f1, double *z, double d1, int *nfgeval);

  // Allocate the memory for the conjugate gradient minimizer
  void CreateConj();

};

#endif
