#include <time.h>
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "incs.h"
#include "getArgs.h"
#include "data.h"
#include "utils.h"
#include "reg_tree.h"
#include "net_regress_mac.h"
#include "net_class_mac.h"
#include "functions.h"

extern "C" int gethostname(char*, ...);

#define DEF_MSTEPS 10
#define DEF_TOL 0.001
#define DEF_ITMAX 10
#define DEF_PRIOR_VAR 0.5
#define DEF_LAMBDA 1
#define FALSE 0
#define TRUE 1
#define DEF_NINP 1
#define DEF_N_OUTPUT 2
#define DEF_N_EXPERTS 2
#define DEF_DEPTH 1
#define DEF_SEED -1
#define DEF_MAX_EPOCHS 100
#define DEF_GATE_RAND 1.0
#define DEF_XPRT_RAND 1.0
#define DEF_NTRAIN 100
#define DEF_DEBUG 0
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 1
#endif


int main(int argc, char **argv) 
{
  char* progName = argv[0];  
  char* version = "0.93";
  getArgs comline = getArgs( &argc, argv );
  // process the command line
  bool more_help = (bool) FALSE;
  if(argc == 1){
    more_help = (bool) TRUE;
  }
  bool help = comline.switchArg( "-help",
				 "\tDisplay command line usage",
				 false );
  int itmax = comline.intArg( "-itmax",
			      "\tMax iterations for optimiser",
			      DEF_ITMAX );
  if((itmax < 1))
    error(1, "itmax = %d : The number of iterations must be at least 1", 
	  itmax);
  int msteps = comline.intArg( "-msteps",
			       "\tMax M steps\t\t",
			       DEF_MSTEPS );
  if((msteps < 1))
    error(1, "msteps = %d : The number of M steps must be at least 1", msteps);

  int rich = comline.intArg( "-rich",
			     "\tRich level for optimiser (0/1)",
			     0 );
  
  if((rich < 0) || (rich > 1))
    error(1, "rich = %d : The richness must be 0 or 1", rich);

  real tol = (real)comline.floatArg( "-tol",
				"\tTolerance level for optimiser",
				DEF_TOL );
  if(tol < 0.0)
    error(1, "tol = %f : The tolerance must be at least 0.0", tol);
  bool anneal = comline.switchArg( "-anneal", 
								   "\tUse annealing",
								   false );

  int debug = comline.intArg( "-debug",
			      "\tdebug level\t\t",
			      DEF_DEBUG );
  int seed = comline.intArg( "-seed",
			     "\trandom seed\t\t",
			     DEF_SEED );
  if(seed < -1)
    error(1, "seed = %d : seed must be greater than 0", seed);
  int n_input = comline.intArg( "-n_input",
				"input vector dimension\t",
				DEF_NINP );
  if(n_input < 1)
    error(1, "n_input = %d : Number of inputs must be > 0", n_input);
  
  int n_output = comline.intArg( "-n_output",
				 "target vector dimension\t",
				 DEF_N_OUTPUT );

  int npoints = comline.intArg( "-n_train",
			       "number of training points",
			       DEF_NTRAIN );
  int n_epochs = comline.intArg( "-n_epochs",
				 "Number of Epochs\t",
				 DEF_MAX_EPOCHS );
  int gate_epochs = comline.intArg( "-gate_epochs",
									"Number of Gate Hyperparameter Epochs\t",
									10 );
  int xprt_epochs = comline.intArg( "-xprt_epochs",
				 "Number of Expert Hyperparameter Epochs\t",
				 10 );
  int n_experts = comline.intArg( "-n_experts",
				  "Number of Experts\t",
				  DEF_N_EXPERTS );
  int depth = comline.intArg( "-depth",
			      "\tDepth of tree\t\t",
			      DEF_DEPTH );

  real lambda = (real)comline.floatArg("-lambda",
				  "\tRegression lambda value\t",
				  DEF_LAMBDA);
  lambda = npoints * 0.1 / pow(n_experts,depth) ;
  printf("lambda = %f\n", lambda);
  real prior_var = (real)comline.floatArg("-prior_var",
				     "Variance prior\t\t",
				     DEF_PRIOR_VAR);
  char* out_file = comline.stringArg( "-out_file",
				      "Weights out file stem\t",
				      "-");
  char* train_file = comline.stringArg( "-train_file",
					"Training data file\t",
					"-");
  char* res_file = comline.stringArg( "-res_file", 
				      "Results file (Matlab format)",
				      "hme_res.m");

  // show command line usage
  if ( help || more_help) {
    comline.printUsage( "test [options]" );
    exit( EXIT_SUCCESS );
  }
  
  comline.finalCheck( 0, "test [options]" );

  if(strcmp(train_file, "-")==0)
    {
      error(0,"You must specify a training data file");
      comline.printUsage( "test [options]" );      
    }
  if(strcmp(out_file, "-")==0)
    {
      error(0,"You must specify an output weights file\n");
      comline.printUsage( "test [options]" );
    }
  char hostname[1024];
  time_t now;
  time(&now);
  assert(gethostname(hostname, 1024)==0);
  printf("Program %s started on %s at %s \n",progName,
		 hostname,ctime(&now));
  printf("Program version = %s\n", version);

  FILE* data_fp;

  data_fp = openFile(train_file, "r");

  Data* train_data_ptr;
  train_data_ptr = new Data(data_fp, npoints, n_input, n_output);
  Data& train_data = *train_data_ptr;

  real** rand_targ = RealMatrix2(npoints, n_output);
  real** rand_in = RealMatrix2(npoints, n_input+1);  
  int* map = new int[npoints];
  if(seed == -1)
    seed = get_seed();

  set_seed(seed);

  Reg_Tree* reg_tree;    

  reg_tree = (Reg_Tree*) new Reg_Tree(npoints, 
									  n_experts, 
									  depth, 
									  n_input, 
									  n_output,
									  debug,
									  0, 0,
									  tol,
									  itmax,
									  msteps,
									  rich,
									  seed,
									  lambda,
									  prior_var,
									  0,0,
									  gate_epochs,
									  xprt_epochs
									  );
  real** orig_inputs = train_data.get_inputs(0);
  real** orig_targets = train_data.get_targets(0);
  real** outputs = RealMatrix2(npoints, n_output);
  real** mom2 = RealMatrix2(npoints, n_output);

  // this section randomizes the inputs and targets
  randomize_matrix_rows2(orig_targets, 
						 npoints, 
						 n_output, map,
						 rand_targ);
  remap_matrix_rows2(orig_inputs, npoints, 
					 n_input+1, map, rand_in);

  real** inputs = rand_in;
  real** targets = rand_targ;

  reg_tree->rand_posts(npoints, inputs, targets);
  //  reg_tree->wta_init(npoints, inputs, targets);

  FILE* fp_res =  openFile(res_file, "w");

  printf("Epoch\tNMSE\tLL\tDiff\tAvg\tJoint\n");
  fprintf(fp_res, "evolv = [\n");
  if(anneal == 1) {
	reg_tree->sa_train(n_epochs, 0, inputs, 
					   targets, fp_res, npoints);
	reg_tree->train(1000, 0, inputs, 
					targets, fp_res, npoints, 1);
  }
  else
	reg_tree->train(n_epochs, 0, inputs, 
					targets, fp_res, npoints, 1);
 
  double trn_err = reg_tree->compute_error(npoints, inputs, targets);
  printf("err = %f\n", trn_err);
  fprintf(fp_res, "];\n");
  reg_tree->write_outputs(orig_inputs, outputs, orig_targets, 
						  mom2, fp_res,
						  npoints);
  fclose(fp_res);
  FILE* out_fp = openFile(out_file, "w");
  reg_tree->write_weights(out_fp);
  fclose(out_fp);
  delete reg_tree;
  time(&now);
  printf("Program %s ended at %s \n",progName,ctime(&now));
}


