#include "incs.h"
#include "net_regress_irls.h"
#include "chol.h"
#include "matrix.h"
#include "utils.h"
#include "functions.h"
#include "lu.h"
#include "svd.h"

NET_Regress_Irls::NET_Regress_Irls(int n_batch, 
				   int num_ip, 
				   int num_op,
				   int dbg,
				   int _msteps,
				   real _tol,
				   real _lambda,
				   real _prior_var  )
  :
  NET_Regress(n_batch, num_ip, num_op, _lambda, _prior_var,
				dbg, _msteps, _tol)
{
  params = RealMatrix2(n_output, n_input);
  grad = RealMatrix2(n_output, n_input);
  grad2 = RealVector(n_output * n_input);
  hessian = RealMatrix3(n_output, n_input, n_input);
  learn_rate = 1.0;
  strcpy(class_str, "NET_Regress_Irls");
}

NET_Regress_Irls::NET_Regress_Irls(const NET_Regress_Irls& old)
  : NET_Regress(old)
{
  params = new real*[n_output];
  for(int i = 0; i < n_output; i ++){
    params[i] = new real[n_input];
  }
  for(int o = 0; o < n_output; o ++) {
	for(int i = 0; i < n_output; i ++){
    params[o][i] = old.params[o][i];
	}
  }
  grad = RealMatrix2(n_output, n_input);
  grad2 = RealVector(n_output * n_input);
  for(int o = 0; o < n_output; o ++) {
	for(int i = 0; i < n_output; i ++){
	  grad[o][i] = old.grad[o][i];
	}
  }
  hessian = RealMatrix3(n_output, n_input, n_input);
  for(int o = 0; o < n_output; o ++) {
	for(int i = 0; i < n_output; i ++){
	  for(int i2 = 0; i2 < n_output; i2 ++){
		hessian[o][i][i2] = old.hessian[o][i][i2];
	  }
	}
  }
  learn_rate = 1.0;
  strcpy(class_str, "NET_Regress_Irls");
}

NET_Regress_Irls::~NET_Regress_Irls()
{
}

double* NET_Regress_Irls::get_grad(int n_points, real** in,
								  real** targ) {
  batch_size = n_points;
  inputs = in;
  targets = targ;
  get_gradient(grad2);
  return grad2;
}

double NET_Regress_Irls::train_params(int n_points)
{

  double cost = 0.0;
  for(int op =0; op < n_output; op ++) {
    for(int ip =0; ip < n_input; ip ++) {
      grad[op][ip] = 0.0;
      for(int ip2 =0; ip2 < n_input; ip2 ++) {
		hessian[op][ip][ip2] = 0.0;
      }
    }
  }
  
  for(int point = 0; point < n_points; point ++){
    for(int op =0; op < n_output; op ++) {
      for(int ip =0; ip < n_input; ip ++) {
		grad[op][ip] += targets[point][op] * inputs[point][ip] * inv_variance[op];
		for(int ip2 =0; ip2 < n_input; ip2 ++) {
		  hessian[op][ip][ip2] +=  inputs[point][ip2] * inputs[point][ip] * inv_variance[op];
		}
      }
    }
  }
  for(int op2 =0; op2 < n_output; op2 ++) {
    lin_solve(hessian[op2], grad[op2], n_input, params[op2]);
  }
  for(int op = 0; op < n_output; op ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      weights[op*n_input + ip] = params[op][ip];
    }
  }
  return cost;
}
/*
double NET_Regress_Irls::train_params(int n_points, 
									  real* wts)
{
  double cost = 0.0;
  for(int op =0; op < n_output; op ++) {
    for(int ip =0; ip < n_input; ip ++) {
      grad[op][ip] = 0.0;
      for(int ip2 =0; ip2 < n_input; ip2 ++) {
		hessian[op][ip][ip2] = 0.0;
      }
    }
  }
  
  for(int point = 0; point < n_points; point ++){
    for(int op =0; op < n_output; op ++) {
      real dink = wts[point] / variance[op];
      for(int ip =0; ip < n_input; ip ++) {
		grad[op][ip] +=  dink * targets[point][op] * inputs[point][ip];	
		for(int ip2 =0; ip2 < n_input; ip2 ++) {
		  hessian[op][ip][ip2] +=  dink * inputs[point][ip2] * inputs[point][ip];
		}
      }
    }
  }
  for(int op2 =0; op2 < n_output; op2 ++) {
    lin_solve(hessian[op2], grad[op2], n_input, params[op2]);
    if(debug > 2){
      printf("params = \n");
      print_vector(params[op2], n_input);
    }
  }
  for(int op = 0; op < n_output; op ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      weights[op*n_input + ip] = params[op][ip];
    }
  }
  return cost;
}
*/
double NET_Regress_Irls::train_params(int n_points, 
									  real* wts)
{
  double cost = 0.0;
  for(int op =0; op < n_output; op ++) {
    for(int ip =0; ip < n_input; ip ++) {
      grad[op][ip] = 0.0;
      for(int ip2 =0; ip2 < n_input; ip2 ++) {
		hessian[op][ip][ip2] = 0.0;
      }
    }
  }
  
  for(int point = 0; point < n_points; point ++){
    for(int op =0; op < n_output; op ++) {
      real dink = wts[point] / variance[op];
      for(int ip =0; ip < n_input; ip ++) {
		grad[op][ip] +=  dink * targets[point][op] * inputs[point][ip];	
		for(int ip2 =0; ip2 < n_input; ip2 ++) {
		  hessian[op][ip][ip2] +=  dink * inputs[point][ip2] * inputs[point][ip];
		}
      }
    }
  }
  for(int op2 =0; op2 < n_output; op2 ++) {
	//    lin_solve(hessian[op2], grad[op2], n_input, params[op2]);
    svd_solve(hessian[op2], grad[op2], params[op2], n_input);
	
    if(debug > 2){
      printf("params = \n");
      print_vector(params[op2], n_input);
    }
  }
  for(int op = 0; op < n_output; op ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      weights[op*n_input + ip] = params[op][ip];
    }
  }
  return cost;
}

