#include "incs.h"
#include "node_nt.h"
#include "node_term.h"
#include "network.h"
#include "utils.h"
#include "functions.h"
#include "net_utils.h"

Node_NT::~Node_NT()
{
  FreeRealVector(&gate_op);
  FreeRealMatrix2(&gate_targets);
}

Node_NT::Node_NT(const Node_NT& old) 
  : Node(old),
    n_child(old.n_child)
{
  child = new Node*[n_child];
  gate_op = RealVector(n_child);
  gate_targets = RealMatrix2(size_batch, n_child);
  strcpy(class_str, "Node_NT");
  for(int n = 0; n < size_batch; n ++) {
	for(int o = 0; o < n_output; o ++) {
	  gate_targets[n][o] = old.gate_targets[n][o];
	}
  }
}


// this constrctor splits terminals into non terminals

Node_NT::Node_NT(Node& old, int num_child) 
  : Node(old),
    n_child(num_child)
{
  child  = new Node*[n_child];

  gate_op = RealVector(n_child);
  gate_targets = RealMatrix2(size_batch, n_child);
  strcpy(class_str, "Node_NT");
}

Node_NT::Node_NT(int n_batch, 
		 int num_child, 
		 int num_ip, 
		 int num_op,
		 Network* gt,
		 Node** xprts, 
		 int dbg,
		 int node_id)
  : Node(n_batch, num_ip, num_op, dbg, node_id),
    n_child(num_child),
    gate(gt),
    child(xprts)
{
  gate_op = RealVector(n_child);
  gate_targets = RealMatrix2(size_batch, n_child);
  strcpy(class_str, "Node_NT");
}

Node_NT::Node_NT(int n_batch, FILE* fp) 
  : Node(n_batch, fp)
{
  strcpy(class_str, "Node_NT");

  fscanf(fp, "%d\n", &id);
  fscanf(fp, "%d\n", &n_child);
  child = new Node*[n_child];
  fscanf(fp, "%d %d\n", &n_input, &n_output);

  gate_op = RealVector(n_child);
  gate_targets = RealMatrix2(size_batch, n_child);

  char* tmp = CharVector(1024);

  for(int ch = 0; ch < n_child; ch ++) {
    fscanf(fp, "%s\n", tmp);
    if(strcmp(tmp,"Node_NT") == 0)
      child[ch] = new Node_NT(size_batch, fp);
    else if(strcmp(tmp,"Node_Term") == 0)
      child[ch] = new Node_Term(size_batch, fp);
    else{
      printf("%s: %s Expected one of Node_NT or Node_Term\n", 
	     class_str, tmp);
      exit(0);
    }
	if(child[ch] == NULL) {
	  error(1, "child %d is null\n", ch);
	}
  }
  fscanf(fp, "%s\n", tmp);
  if(strcmp(tmp, "Gate")!=0){
    printf("%s : I was expecting a gate here...\n", class_str);
    exit(0);
  }
  fscanf(fp, "%s\n", tmp);
  if(strstr(tmp, "NET_Class") == NULL) {
    printf("%s : Expected a NET_Class here\n", class_str);
    exit(0);
  }
  else{
	gate = get_new_net(size_batch, fp, tmp);
  }
  if(gate == NULL) {
	error(1, "Gate is NULL !!");
  }
  FreeCharVector(&tmp);
}

// ---------------------------------------------------------
// 
// ---------------------------------------------------------

void Node_NT::set_joint_prob(real prob, real* in, int point) {
  joint_prob[point] = prob;
  gate->go_forward_log(in, gate_op);

  for(int ch = 0; ch < n_child; ch ++) {
	child[ch]->set_cond_prob(point, gate_op[ch]);
	child[ch]->set_joint_prob(prob + gate_op[ch], in, point);
  }
}

void Node_NT::train_gates(int n_points, real** in)
{
  if(test_prune(joint_post, n_points) == 1) {
	//	printf("pruning out\n");
	return;
  }
  get_new_gate_targets(n_points);
  if(n_child > 1){
	for(int i = 0; i < n_points; i ++)
	  if(debug >2) check_arg(joint_post[i], "node_nt jp");
	gate->train(n_points, in, gate_targets, joint_post);
  }
  for(int xp = 0; xp < n_child; xp ++) {
	if(strcmp(child[xp]->get_class(), "Node_NT") ==0) {
	  Node_NT* ch = (Node_NT*) child[xp];
	  ch->train_gates(n_points, in);
	}
  }
  
}

// ---------------------------------------------------------
// 
// ---------------------------------------------------------

void Node_NT::update_params(int n_points, real** in, 
							real** targ)
{
  if(test_prune(joint_post, n_points) == 1) {
	//	printf("pruning out\n");
	return;
  }
  if(n_child > 1){
	get_new_gate_targets(n_points);
	gate->train(n_points, in, gate_targets, joint_post);
  }
  for(int xp = 0; xp < n_child; xp ++) {
	child[xp]->update_params(n_points, in, targ);
  }
}

void Node_NT::update_hypers(int n_points, real** in, real** targ)
{
  if(test_prune(joint_post, n_points) == 1) {
	//	printf("pruning out\n");
	return;
  }
  if(n_child > 1){
	get_new_gate_targets(n_points);
	gate->update_hyperparameters(n_points, 
								 in, 
								 gate_targets,
								 joint_post);
  }
  for(int xp = 0; xp < n_child; xp ++) {
	child[xp]->update_hypers(n_points, in, targ);
  }
}

void Node_NT::train(int n_points, real** in, real** targ)
{
  if(test_prune(joint_post, n_points) == 1) {
	//	printf("pruning out\n");
	return;
  }
  if(n_child > 1){
	get_new_gate_targets(n_points);
	gate->train(n_points, in, gate_targets, joint_post);
  }
  for(int xp = 0; xp < n_child; xp ++) {
	child[xp]->train(n_points, in,  targ);
  }
}

// ---------------------------------------------------------
// 
// ---------------------------------------------------------

void Node_NT::find_posts2(int point) {
  // given joint posteriors of children, what is conditional
  if(visited == 1)
	return;
  real sum = LZERO;
  real jp;
  for(int ch1 = 0; ch1 < n_child; ch1 ++) {
	jp = child[ch1]->get_joint_post(point);
	if(debug >2) check_arg(jp, "find_posts:jp");
	if(jp > LSMALL)
	  sum = LAdd(sum, jp);
  }
  for(int ch = 0; ch < n_child; ch ++) {
	double cp = child[ch]->get_joint_post(point) - sum;
	child[ch]->set_cond_post(point, cp);
	if(debug >2) check_arg(cp, "find_posts:cp");
  }
  joint_post[point] = sum;
  if(debug >2) check_arg(joint_post[point], "find_posts:jp2");
  visited = 1;
  if(parent == NULL) {
	cond_post[point] = 0.0;
  }
  else{
	Node_NT* pa = (Node_NT*) parent;
	pa->find_posts2(point);
  }
}

void Node_NT::find_posts1(int point) {
  // given joint posteriors of children, what is conditional
  if(visited == 1)
	return;
  real sum = LZERO;
  real jp;
  for(int ch1 = 0; ch1 < n_child; ch1 ++) {
	jp = child[ch1]->get_joint_post(point);
	if(debug >2) check_arg(jp, "find_posts:jp");
	if(jp > LSMALL)
	  sum = LAdd(sum, jp);
  }
  for(int ch = 0; ch < n_child; ch ++) {
	double cp = child[ch]->get_joint_post(point) - sum;
	child[ch]->set_cond_post(point, cp);
	if(debug >2) check_arg(cp, "find_posts:cp");
  }
  joint_post[point] = sum;
  if(debug >2) check_arg(joint_post[point], "find_posts:jp2");
  visited = 1;
}

void Node_NT::reset_nodes() {
  visited = 0;
  for(int xp = 0; xp < n_child; xp ++)
	child[xp]->reset_nodes();
}
// ---------------------------------------------------------
// 
// ---------------------------------------------------------
void Node_NT::randomize_weights(real xprt_offset, real gate_offset) 
{
  for(int xp = 0; xp < n_child; xp ++) {
    child[xp]->randomize_weights(xprt_offset, gate_offset);
  }
  gate->randomize_weights(gate_offset);
}

void Node_NT::randomize_weights(real offset) 
{
  for(int xp = 0; xp < n_child; xp ++) {
    child[xp]->randomize_weights(offset);
  }
  gate->randomize_weights(offset);
}
// ---------------------------------------------------------
// 
// ---------------------------------------------------------

void Node_NT::print_posteriors(int point)
{
  printf("Node_NT: point %d cp=%f jp=%f cpr=%f jpr=%f\n", 
	 point, Sexp(cond_post[point]), Sexp(joint_post[point]),
	 Sexp(cond_prob[point]), Sexp(joint_prob[point]));
  for(int xp = 0; xp < n_child; xp ++){
    child[xp]->print_posteriors(point);
  }
}

void Node_NT::write_weights(FILE* fp)
{
  fprintf(fp, "%s\n", class_str);
  fprintf(fp, "%d\n", id);
  fprintf(fp, "%d\n", n_child);
  fprintf(fp, "%d %d\n", n_input, n_output);
  for(int xp = 0; xp < n_child; xp ++){
    child[xp]->write_weights(fp);
  }
  fprintf(fp, "Gate\n");
  gate->write_weights(fp);
}

void Node_NT::read_weights(FILE* fp)
{
  fscanf(fp, "%s\n", class_str);
  fscanf(fp, "%d\n", &id);
  fscanf(fp, "%d\n", &n_child);
  fscanf(fp, "%d %d\n", &n_input, &n_output);
  for(int xp = 0; xp < n_child; xp ++){
    child[xp]->read_weights(fp);
  }
  char tmp[1024];
  fscanf(fp, "%s\n", tmp);
  if(strcmp(tmp, "Gate")!=0)
    error(1, "I was expecting a gate here...\n");
  gate->read_weights(fp);
}

void Node_NT::init(int n_points, real** in, 
				   real** targ) {
  get_new_gate_targets(n_points);
  gate->init(n_points, in, gate_targets, joint_post);
  for(int xp = 0; xp < n_child; xp ++) {
	child[xp]->init(n_points, in, targ);
  }
}
void Node_NT::get_new_gate_targets(int n_points) {

  for(int point = 0; point < n_points; point ++){
    for(int xp = 0; xp < n_child; xp ++){
	  real cp = Sexp(child[xp]->get_cond_post(point));
	  if(cp < VSMALL) cp = VSMALL;
      gate_targets[point][xp] = cp;
	  
	  if(debug >2) check_arg(cp, "cond_post");
    }
  }
}

void Node_NT::replace_child(int node_id, Node* ch) {
  int repl_fl = 0;
  for(int xp = 0; xp < n_child; xp ++) {
	if(child[xp]->get_id() == node_id) {
	  child[xp] = ch;
	  repl_fl = 1;
	  break;
	}
  }
  if(repl_fl == 0)
	printf("didnt replace in this fn\n");
}

real Node_NT::get_dens() {
  return prob_dens;
}

real Node_NT::set_dens(int point, real* in, real* targ) {
  real dens = 0.0;
  for(int xp =0; xp < n_child; xp ++) {
	dens = LAdd(dens, child[xp]->get_cond_prob(point) + child[xp]->set_dens(point, in, targ));
  }
  prob_dens = dens;
  return dens;
}

void Node_NT::set_posts(int point) {
  if(parent == NULL) {
	joint_post[point] = 0.0;
	cond_post[point] = 0.0;
  }
  real sum = LZERO;
  
  for(int xp =0; xp < n_child; xp ++) {
	real tmp = child[xp]->get_dens() + child[xp]->get_cond_prob(point);
	child[xp]->set_cond_post(point, tmp);
	sum = LAdd(sum, tmp);
  }
  for(int xp =0; xp < n_child; xp ++) {
	real tmp = child[xp]->get_cond_post(point);
	child[xp]->set_cond_post(point, tmp - sum);
  }
  for(int xp =0; xp < n_child; xp ++) {
	real cp = child[xp]->get_cond_post(point);
	child[xp]->set_joint_post(point, cp + joint_post[point]);
  }
  for(int xp =0; xp < n_child; xp ++) {
	child[xp]->set_posts(point);
  }
}

