#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include "utils.h"
#include "incs.h"

/*---------------------------------------------------------*/
/* FUNCTION -> openFile                                    */
/* Open a file and check that it can be opened.            */
/* Exits if fails, returns file stream otherwise           */
/*---------------------------------------------------------*/
FILE *openFile(const char *filename,
	       const char *mode,
	       char *errMsg) {

  FILE *fp = fopen(filename, mode);
  if (fp == NULL) {
    error(1, "%s\nUnable to open file %s (with mode %s)\n", 
	     errMsg, filename, mode);
    exit(1);
  }

  return fp;

}

/*---------------------------------------------------------*/
/* FUNCTION -> closeFile                                   */
/* close a file: exits if there's an error                 */
/*---------------------------------------------------------*/
void closeFile(FILE *stream,
	       char *errMsg) {

  if (fclose(stream) == EOF) {
    error(1, "%s\nUnable to close file stream properly\n", errMsg);
    exit(1);
  }
  
}

/*---------------------------------------------------------*/
/* FUNCTION -> bfread                                      */
/* Do a binary read using fread and check for errors       */
/* returns zero if EOF found                               */
/*---------------------------------------------------------*/
int bfread( void *ptr, 
	       int size, 
	       int nobj, 
	       FILE *stream,
	       char *errMsg) {

  // first check pointer and stream are not NULL pointers
  if (stream == NULL)
    error(1, "%s\nStream is a NULL pointer!\n", errMsg);

  if (ptr == NULL)
    error(1, "%s\nReading data into a NULL pointer!\n", errMsg);
    
  int itemsRead = fread(ptr, size, nobj, stream);

  // checking items read
  if (itemsRead < nobj && feof(stream) == 0)
    error(1, "%s\nOnly %d of %d requested items read from binary file\n",
	     errMsg, itemsRead, nobj);

  // checking for errors
  if (ferror(stream) != 0)
    error(1, "%s\nProblems reading from binary file\n", errMsg);

  // checking for file end of file
  if (feof(stream) != 0)
    return 0;
  else
    return 1;

}

/*---------------------------------------------------------*/
/* FUNCTION -> bfread                                      */
/* Do a binary read using fread and check for errors       */
/* returns great than zero if EOF found                    */
/*---------------------------------------------------------*/
void bfwrite( const void *ptr,
	      int size, 
	      int nobj, 
	      FILE *stream,
	      char *errMsg) {

  // first check pointer and stream are not NULL pointers
  if (stream == NULL)
    error(1, "%s\nOutputting to a NULL pointer stream!\n", errMsg);

  if (ptr == NULL)
    error(1, "%s\nWriting data from a NULL pointer!\n", errMsg);

  int itemsWritten = fwrite(ptr, size, nobj, stream);

  // checking items written
  if (itemsWritten < nobj)
    error(1, "%s\nOnly %d of %d requested items written to binary file\n",
	     errMsg, itemsWritten, nobj);

}

/*---------------------------------------------------------*/
/* FUNCTION -> error	                                   */
/* Prints toolname and then message to stderr.  Form of    */
/* message and any subsequent arguments is similar to      */
/* 'printf' with just %d %f %e %s being recognised.  If    */
/* status<>0 then tool is aborted and status returned to   */
/* OS, return status is also printed in error message in   */
/* this case.                                              */
/*---------------------------------------------------------*/
void error(int errcode, 
	      char *message, 
	      ...) {

  va_list ap;              /* Pointer to unnamed args */
  char *p;                /* Message String */
  double dval;            /* Message args */
  int ival,cval;
  char *sval;
  FILE *f;
   
   fflush(stdout);        /* Flush any pending output */
   va_start(ap,message);
   if (errcode<=0) {
     fprintf(stdout,"WARNING ");
     f = stdout;
   }else{
     fprintf(stderr,"ERROR ");
     f = stderr;
   }
   for (p=message;*p;p++){
     if (*p != '%'){
       fputc(*p,f);
       continue;
     }
     switch(*++p){
     case 'd':
       ival = va_arg(ap,int);
       fprintf(f,"%d",ival);
       break;
     case 'f':
       dval = va_arg(ap,double);
       fprintf(f,"%f",dval);
       break;
     case 'e':
       dval = va_arg(ap,double);
       fprintf(f,"%e",dval);
       break;      
     case 's':
       sval = va_arg(ap,char *);
       fprintf(f,"%s",sval);
       break;
     case 'c':
       cval = va_arg(ap,int);
       fprintf(f,"%c",cval);
     }
   }
   va_end(ap);
   if (errcode>0) {
     fprintf(f,"\nABORTED [%d]\n",errcode);
     exit(errcode);    
   } else {
     fprintf(f,"\n");
     fflush(f);
   }
 }

/* call malloc() with error checking */

void *Smalloc(int nbyte) {
  void *address;

  if(nbyte == 0) nbyte = 1;

  if((address = malloc(nbyte)) == NULL)
    error(1,"Smalloc(%d) == NULL", nbyte);
  return(address);
}

/* call Sfree() with error checking */

void Sfree(void *address) {
  if(address == NULL) error(1,"Sfree(NULL)\n");
  free(address);
#ifdef FREE_RETURNS_STATUS
  if((char*) free(address) == NULL)
    error(1,"Sfree(%x) == NULL\n", (int) address);
#endif
}

/* Float routines here */

float *floatvector(int size) {
  return((float*) malloc( (size * sizeof(float))));
}
float *FloatVector(int size) {
  float *array = floatvector(size);
  if(array == NULL) error(1,"FloatVector(%d) == NULL\n", size);
  for(int i = 0; i < size; i ++)
	array[i] = 0.0;
  return(array);
}

void FreeFloatVector(float **address)
{
  Sfree(*address);
  *address = (float*)NULL;
}

float **floatmatrix2(int size0, int size1) {
  float** array0 = NULL;
  float*  array1 = NULL;
  if(((array0 = (float**) malloc(size0 * sizeof(float*))) != NULL) &&
     ((array1 = (float*)  malloc(size0 * size1* sizeof(float))) != NULL)){
    int i;
    for(i = 0; i < size0; i++)
      array0[i] = array1 + i * size1;
  }
  return(array0);
}

float ***floatmatrix3(int size0, int size1, int size2) {
  float *** array0 = NULL;
  float  **array1 = NULL;
  float   *array2 = NULL;
  if(((array0 = (float***) malloc(size0 * sizeof(float**))) != NULL) &&
     ((array1 = (float**)  malloc(size0 * size1* sizeof(float*))) != NULL) &&
     ((array2 = (float*)  malloc(size0 * size1* size2* sizeof(float))) != NULL)){
    int i,j;
    for(i = 0; i < size0; i++){
      array0[i] = array1 +i *size1;
      for(j = 0; j < size1; j++)
	array0[i][j] = array2 + (i * size1 + j) * size2;
    }
  }
  return(array0);
}

float **FloatMatrix2(int i_size, int j_size) {
  float **array = floatmatrix2(i_size, j_size);
  if(array == NULL)
    error(1,"FloatMatrix2(%d, %d) == NULL\n", i_size, j_size);
  for(int i = 0; i < i_size; i ++)
	for(int j = 0; j < j_size; j ++)
	  array[i][j] = 0.0;
  return(array);
}

void FreeFloatMatrix2(float ***address)
{
  Sfree(**address);
  Sfree(*address);
  *address = (float**) NULL;
}

float ***FloatMatrix3(int i_size, int j_size, int k_size) {
  float ***array = floatmatrix3(i_size, j_size, k_size);
  if(array == NULL)
    error(1,"FloatMatrix3(%d, %d, %d) == NULL\n", i_size,j_size,k_size);
  for(int i = 0; i < i_size; i ++)
	for(int j = 0; j < j_size; j ++)
	  for(int k = 0; k < k_size; k ++)
		array[i][j][k] = 0;

  return(array);
}


void FreeFloatMatrix3(float ****address)
{
  Sfree(***address);
  Sfree(**address);
  Sfree(*address);
  *address = (float***)NULL;
}


int **intmatrix2(int size0, int size1) {
  int** array0 = NULL;
  int*  array1 = NULL;
  if(((array0 = (int**) malloc(size0 * sizeof(int*))) != NULL) &&
     ((array1 = (int*)  malloc(size0 * size1* sizeof(int))) != NULL)){
    int i;
    for(i = 0; i < size0; i++)
      array0[i] = array1 + i * size1;
  }
  return(array0);
}

int **IntMatrix2(int i_size, int j_size) {
  int **array = intmatrix2(i_size, j_size);
  if(array == NULL)
    error(1,"intMatrix2(%d, %d) == NULL\n", i_size, j_size);
  for(int i = 0; i < i_size; i ++)
	for(int j = 0; j < j_size; j ++)
	  array[i][j] = 0;

  return(array);
}

/* int rountines here */

int *intvector(int size) {
  return((int*) malloc((unsigned) (size * sizeof(int))));
}
int *IntVector(int size) {
  int *array = intvector(size);
  if(array == NULL) error(1,"SintVector(%d) == NULL\n", size);
  for(int i = 0; i < size; i ++)
	array[i] = 0;
  return(array);
}
void FreeIntVector(int **address)
{
  Sfree(*address);
  *address = (int*)NULL;
}
/* char routines here */

char *CharVector(int size) {
  char *array = charvector(size);
  if(array == NULL) error(1,"Scharvector(%d) == NULL\n", size);
  return(array);
}
char *charvector(int size) {
  return((char*) malloc((unsigned) (size * sizeof(char))));
}
 
unsigned char *UcharVector(int size) {
  unsigned char *array = ucharvector(size);
  if(array == NULL) error(1,"Scharvector(%d) == NULL\n", size);
  return(array);
}
unsigned char *ucharvector(int size) {
  return((unsigned char*) malloc((unsigned) (size * sizeof(unsigned char))));
}
void FreeCharVector(char **address)
{
  Sfree(*address);
  *address = (char*)NULL;
} 

/* Double routines here */

double *doublevector(int size) {
  return((double*) malloc( (size * sizeof(double))));
}
double *DoubleVector(int size) {
  double *array = doublevector(size);
  if(array == NULL) error(1,"DoubleVector(%d) == NULL\n", size);
  for(int i = 0; i < size; i ++) 
	array[i] = 0.0;
  return(array);
}

void FreeDoubleVector(double **address)
{
  Sfree(*address);
  *address = (double*)NULL;
}

double **doublematrix2(int size0, int size1) {
  double** array0 = NULL;
  double*  array1 = NULL;
  if(((array0 = (double**) malloc(size0 * sizeof(double*))) != NULL) &&
     ((array1 = (double*)  malloc(size0 * size1* sizeof(double))) != NULL)){
    int i;
    for(i = 0; i < size0; i++)
      array0[i] = array1 + i * size1;
  }
  return(array0);
}

double ***doublematrix3(int size0, int size1, int size2) {
  double *** array0 = NULL;
  double  **array1 = NULL;
  double   *array2 = NULL;
  if(((array0 = (double***) malloc(size0 * sizeof(double**))) != NULL) &&
     ((array1 = (double**)  malloc(size0 * size1* sizeof(double*))) != NULL) &&
     ((array2 = (double*)  malloc(size0 * size1* size2* sizeof(double))) != NULL)){
    int i,j;
    for(i = 0; i < size0; i++){
      array0[i] = array1 +i *size1;
      for(j = 0; j < size1; j++)
	array0[i][j] = array2 + (i * size1 + j) * size2;
    }
  }
  return(array0);
}

double **DoubleMatrix2(int i_size, int j_size) {
  double **array = doublematrix2(i_size, j_size);
  if(array == NULL)
    error(1,"DoubleMatrix2(%d, %d) == NULL\n", i_size, j_size);
  for(int i = 0; i < i_size; i ++)
	for(int j = 0; j < j_size; j ++)
	  array[i][j] = 0.0;

  return(array);
}

void FreeDoubleMatrix2(double ***address)
{
  Sfree(**address);
  Sfree(*address);
  *address = (double**) NULL;
}

double ***DoubleMatrix3(int i_size, int j_size, int k_size) {
  double ***array = doublematrix3(i_size, j_size, k_size);
  if(array == NULL)
    error(1,"DoubleMatrix3(%d, %d, %d) == NULL\n", i_size,j_size,k_size);
  for(int i = 0; i < i_size; i ++)
	for(int j = 0; j < j_size; j ++)
	  for(int k = 0; k < k_size; k ++)
		array[i][j][k] = 0;

  return(array);
}


void FreeDoubleMatrix3(double ****address)
{
  Sfree(***address);
  Sfree(**address);
  Sfree(*address);
  *address = (double***)NULL;
}





/* Real routines here */

real *realvector(int size) {
  return((real*) malloc( (size * sizeof(real))));
}
real *RealVector(int size) {
  real *array = realvector(size);
  if(array == NULL) error(1,"RealVector(%d) == NULL\n", size);
  for(int i = 0; i < size; i ++) 
	array[i] = 0.0;
  return(array);
}

void FreeRealVector(real **address)
{
  Sfree(*address);
  *address = (real*)NULL;
}

real **realmatrix2(int size0, int size1) {
  real** array0 = NULL;
  real*  array1 = NULL;
  if(((array0 = (real**) malloc(size0 * sizeof(real*))) != NULL) &&
     ((array1 = (real*)  malloc(size0 * size1* sizeof(real))) != NULL)){
    int i;
    for(i = 0; i < size0; i++)
      array0[i] = array1 + i * size1;
  }
  return(array0);
}

real ***realmatrix3(int size0, int size1, int size2) {
  real *** array0 = NULL;
  real  **array1 = NULL;
  real   *array2 = NULL;
  if(((array0 = (real***) malloc(size0 * sizeof(real**))) != NULL) &&
     ((array1 = (real**)  malloc(size0 * size1* sizeof(real*))) != NULL) &&
     ((array2 = (real*)  malloc(size0 * size1* size2* sizeof(real))) != NULL)){
    int i,j;
    for(i = 0; i < size0; i++){
      array0[i] = array1 +i *size1;
      for(j = 0; j < size1; j++)
	array0[i][j] = array2 + (i * size1 + j) * size2;
    }
  }
  return(array0);
}

real **RealMatrix2(int i_size, int j_size) {
  real **array = realmatrix2(i_size, j_size);
  if(array == NULL)
    error(1,"RealMatrix2(%d, %d) == NULL\n", i_size, j_size);
  for(int i = 0; i < i_size; i ++)
	for(int j = 0; j < j_size; j ++)
	  array[i][j] = 0.0;

  return(array);
}

void FreeRealMatrix2(real ***address)
{
  Sfree(**address);
  Sfree(*address);
  *address = (real**) NULL;
}

real ***RealMatrix3(int i_size, int j_size, int k_size) {
  real ***array = realmatrix3(i_size, j_size, k_size);
  if(array == NULL)
    error(1,"RealMatrix3(%d, %d, %d) == NULL\n", i_size,j_size,k_size);
  for(int i = 0; i < i_size; i ++)
	for(int j = 0; j < j_size; j ++)
	  for(int k = 0; k < k_size; k ++)
		array[i][j][k] = 0;

  return(array);
}


void FreeRealMatrix3(real ****address)
{
  Sfree(***address);
  Sfree(**address);
  Sfree(*address);
  *address = (real***)NULL;
}
